"""
Main web service CLI

"""
from click import command, option
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.app_hooks import create_serve_app
from microcosm_sagemaker.artifact import InputArtifact
from microcosm_sagemaker.commands.config import load_default_runserver_config
from microcosm_sagemaker.commands.options import input_artifact_option


@command()
@option(
    "--host",
    default="127.0.0.1",
)
@option(
    "--port",
    type=int,
)
@option(
    "--debug/--no-debug",
    default=False,
)
@input_artifact_option()
def main(host, port, debug, input_artifact):
    graph = create_serve_app(
        debug=debug,
        loaders=[
            load_default_runserver_config,
            input_artifact.load_config,
        ],
    )

    run_serve(
        graph=graph,
        input_artifact=input_artifact,
        host=host,
        port=port,
    )


def run_serve(
        graph: ObjectGraph,
        input_artifact: InputArtifact,
        host: str,
        port: int,
) -> None:
    graph.active_bundle.load(input_artifact)

    graph.flask.run(
        host=host,
        port=port or graph.config.flask.port,
    )



"""
Ping controller.
"""
from marshmallow import Schema, fields
from microcosm_flask.conventions.base import Convention
from microcosm_flask.conventions.encoding import dump_response_data
from microcosm_flask.conventions.registry import response
from microcosm_flask.namespaces import Namespace
from microcosm_flask.operations import Operation


class PingSchema(Schema):
    """
    AWS health check.

    """
    status = fields.String()


class PingConvention(Convention):
    def configure_query(self, ns, definition):
        @self.add_route("/ping", Operation.Query, ns)
        @response(PingSchema())
        def ping_query():
            # NB: We have no internal consistency check, but we'll likely want to
            # add this & change the response status code accordingly
            # XXX At some point we may want to add the ability to register checks,
            # similar to how we do for the microcosm-flask health convention.
            response_data = dict(
                status="healthy",
            )
            return dump_response_data(PingSchema(), response_data, 200)


def configure_ping(graph):
    ns = Namespace(
        subject="ping",
        version=None
    )
    convention = PingConvention(graph)

    convention.configure(ns, query=tuple())



"""
Main training CLI

"""
import json

from click import File, command, option
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.app_hooks import create_train_app
from microcosm_sagemaker.artifact import OutputArtifact
from microcosm_sagemaker.commands.options import input_data_option, output_artifact_option
from microcosm_sagemaker.exceptions import raise_sagemaker_exception
from microcosm_sagemaker.input_data import InputData


@command()
@option(
    "--configuration",
    type=File('r'),
    help="Manual import of configuration file, used for local testing",
)
@input_data_option()
@output_artifact_option()
@option(
    "--auto-evaluate/--no-auto-evaluate",
    default=True,
    help="Whether to automatically evaluate after the training has completed",
)
def main(configuration, input_data, output_artifact, auto_evaluate):
    try:
        extra_config = json.load(configuration) if configuration else dict()

        graph = create_train_app(extra_config=extra_config)

        run_train(graph, input_data, output_artifact)

        if auto_evaluate:
            run_auto_evaluate(graph, input_data)
    except Exception as e:
        raise_sagemaker_exception(e)


def run_train(
    graph: ObjectGraph,
    input_data: InputData,
    output_artifact: OutputArtifact,
) -> None:
    output_artifact.init()
    output_artifact.save_config(graph.config)

    graph.training_initializers.init()

    bundle = graph.active_bundle
    bundle.fit(input_data)
    bundle.save(output_artifact)


def run_auto_evaluate(graph: ObjectGraph, input_data: InputData) -> None:
    graph.active_evaluation(graph.active_bundle, input_data)



import json
from typing import Callable, NamedTuple, TypeVar

from hamcrest.core.base_matcher import BaseMatcher


T = TypeVar('T')


class ExtractorMatcherPair(NamedTuple):
    """
    Contains a pair of extractor and matcher which is used during testing.  The
    extractor will be applied to the raw bytes of a file output during testing,
    and then the matcher will be run on the output of the extractor.

    """
    # NB: We cannot use a dataclass due to python/mypy#5485
    extractor: Callable[[bytes], T]
    matcher_constructor: Callable[[T], BaseMatcher]


def json_extractor(raw_bytes):
    return json.loads(raw_bytes.decode())



import tempfile
from pathlib import Path
from typing import Mapping

from microcosm_sagemaker.commands import train
from microcosm_sagemaker.testing.bytes_extractor import ExtractorMatcherPair
from microcosm_sagemaker.testing.cli_test_case import CliTestCase
from microcosm_sagemaker.testing.directory_comparison import directory_comparison


class TrainCliTestCase(CliTestCase):
    """
    Helper base class for writing tests of the train cli.

    """
    def test_train(
        self,
        input_data_path: Path,
        gold_output_artifact_path: Path,
        output_artifact_matchers: Mapping[Path, ExtractorMatcherPair],
    ) -> None:
        """
        Runs the `train` command on the given `input_data_path` and then
        recursively checks the contents of the output artifact against the
        expected contents in `gold_output_artifact_path`.  It is also possible
        to leave certain files out of the gold dir, and instead specify a
        matcher that should be used for the contents of the given file instead.

        """
        with tempfile.TemporaryDirectory() as output_artifact_path:
            self.run_and_check(
                command_name="train",
                command=train.main,
                args=[
                    "--input-data",
                    str(input_data_path),
                    "--output-artifact",
                    output_artifact_path,
                ],
            )

            directory_comparison(
                gold_dir=gold_output_artifact_path,
                actual_dir=Path(output_artifact_path),
                matchers=output_artifact_matchers,
            )



from random import seed

from microcosm.api import defaults


@defaults(
    seed=42,
)
class Random:
    def __init__(self, graph):
        self.seed = graph.config.random.seed

        graph.training_initializers.register(self)

    def init(self):
        seed(self.seed)



from abc import ABC, abstractmethod
from typing import Any

from microcosm_sagemaker.artifact import InputArtifact, OutputArtifact
from microcosm_sagemaker.input_data import InputData


class Bundle(ABC):
    @abstractmethod
    def fit(self, input_data: InputData) -> None:
        """
        Perform training

        """
        pass

    @abstractmethod
    def predict(self) -> Any:
        """
        Predict using the trained model

        Note that derived classes can define their own parameters and are
        expected to return something.

        """
        pass

    @abstractmethod
    def save(self, output_artifact: OutputArtifact) -> None:
        """
        Save the trained model

        """
        pass

    @abstractmethod
    def load(self, input_artifact: InputArtifact) -> None:
        """
        Load the trained model

        """
        pass



"""
Consumer factories.

"""


def configure_active_bundle(graph):
    if not getattr(graph.config, "active_bundle", ""):
        return None
    return getattr(graph, graph.config.active_bundle)


def configure_active_evaluation(graph):
    if not getattr(graph.config, "active_evaluation", ""):
        return None
    return getattr(graph, graph.config.active_evaluation)


def configure_sagemaker(graph):
    """
    Instantiates all the necessary sagemaker factories.

    """
    graph.use(
        "active_bundle",
        "active_evaluation",
        "training_initializers",
        "random",
    )



"""
Loaders to inject SM parameters as microcosm configurations.

"""
import json

from boto3 import client
from microcosm.config.model import Configuration
from microcosm.loaders.compose import merge
from microcosm.loaders.keys import expand_config
from microcosm.metadata import Metadata

from microcosm_sagemaker.constants import SagemakerPath
from microcosm_sagemaker.s3 import S3Object


def load_from_hyperparameters(metadata: Metadata) -> Configuration:
    """
    Sagemaker only supports single-layer hyperparameters, so we use double underscores
    (__) to signify the delineation between nested dictionaries.  This mirrors the
    formatting of our ENV variables.  Note that these values are all strings by convention,
    so any end applications should.

    This configuration helper parses these into the underlying dictionary structure.

    """
    try:
        with open(SagemakerPath.HYPERPARAMETERS) as raw_file:
            return expand_config(
                json.load(raw_file),
                separator="__",
                skip_to=0,
            )

    except FileNotFoundError:
        return Configuration()


def load_from_s3(url: str) -> Configuration:
    """
    Loads a S3 formatted url that points to a remote json file, and parses it into a local
    configuration variable.

    """
    def _load(metadata):
        s3 = client("s3")
        s3_object = S3Object.from_url(url)

        object = s3.get_object(Bucket=s3_object.bucket, Key=s3_object.key)
        return Configuration(json.loads(object["Body"].read()))

    return _load


def load_train_conventions(metadata: Metadata) -> Configuration:
    """
    Opinionated loader that:
    1. Reads all of the hyperparameters passed through by SageMaker
    2. Uses a special `base_configuration` key to read the given configuration from S3

    """
    configuration = load_from_hyperparameters(metadata)
    base_configuration_url = configuration.pop("base_configuration", None)

    if base_configuration_url:
        remote_configuration = load_from_s3(base_configuration_url)(metadata)

        # Locally specified hyperparameters should take precedence over the
        # base configuration
        configuration = merge([
            remote_configuration,
            configuration,
        ])

    return configuration


def load_serve_conventions(metadata: Metadata) -> Configuration:
    """
    Loads the configuration.json file from the input artifact directory on the docker

    """
    model_path = SagemakerPath.MODEL
    config_path = model_path / "configuration.json"

    try:
        with open(config_path) as raw_file:
            return expand_config(
                json.load(raw_file),
                skip_to=0,
            )

    except FileNotFoundError:
        return Configuration()



from pathlib import Path
from typing import Mapping, Optional

from hamcrest import (
    assert_that,
    contains,
    equal_to,
    is_,
)

from microcosm_sagemaker.testing.bytes_extractor import ExtractorMatcherPair


def identity(x):
    return x


def directory_comparison(
    gold_dir: Path,
    actual_dir: Path,
    matchers: Optional[Mapping[Path, ExtractorMatcherPair]] = None,
):
    """
    Recursively checks the contents of `actual_dir` against the expected
    contents in `gold_dir`.  It is also possible to leave certain files out of
    the gold dir, and instead specify an (extractor, matcher) pair that should
    be used to extract and match the contents of the given file instead.

    """
    matchers = matchers or dict()

    actual_paths = sorted([
        subpath.relative_to(actual_dir)
        for subpath in actual_dir.glob('**/*')
    ])
    gold_paths = sorted([
        subpath.relative_to(gold_dir)
        for subpath in gold_dir.glob('**/*')
    ])

    assert_that(actual_paths, contains(*gold_paths))

    for path in gold_paths:
        gold_path = gold_dir / path
        actual_path = actual_dir / path

        if gold_path.is_dir():
            assert_that(actual_path.is_dir(), is_(True))
        else:
            if path in matchers:
                extractor, matcher_constructor = matchers[path]
            else:
                extractor, matcher_constructor = ExtractorMatcherPair(
                    identity,
                    lambda x: is_(equal_to(x)),
                )

            assert_that(
                extractor(actual_path.read_bytes()),
                matcher_constructor(
                    extractor(gold_path.read_bytes()),
                ),
                path,
            )



"""
Example CRUD routes tests.

Tests are sunny day cases under the assumption that framework conventions
handle most error conditions.

"""
from pathlib import Path

from hamcrest import (
    assert_that,
    equal_to,
    has_entries,
    is_,
)
from hamcrest.core.base_matcher import BaseMatcher

from microcosm_sagemaker.app_hooks import create_serve_app
from microcosm_sagemaker.artifact import InputArtifact
from microcosm_sagemaker.commands.config import load_default_runserver_config


class InvocationsRouteTestCase:
    """
    Helper base class for writing tests of the invocations route.

    """
    def setup(self, input_artifact_path: Path) -> None:
        self.input_artifact = InputArtifact(input_artifact_path)

        self.graph = create_serve_app(
            testing=True,
            loaders=[
                load_default_runserver_config,
                self.input_artifact.load_config,
            ],
        )

        self.client = self.graph.flask.test_client()

    def test_search(
        self,
        request_json: dict,
        response_items_matcher: BaseMatcher,
    ) -> None:
        """
        Invokes the invocations endpoint with `request_json`, and checks the
        `items` entry of the response against `response_items_matcher`.

        """
        self.graph.active_bundle.load(self.input_artifact)

        uri = "/invocations"

        response = self.client.post(
            uri,
            json=request_json,
        )

        assert_that(response.status_code, is_(equal_to(200)))
        assert_that(
            response.json,
            has_entries(
                items=response_items_matcher,
            ),
        )



"""
Main evaluation CLI

"""
from click import command
from microcosm.object_graph import ObjectGraph

from microcosm_sagemaker.app_hooks import create_evaluate_app
from microcosm_sagemaker.artifact import InputArtifact
from microcosm_sagemaker.commands.options import input_artifact_option, input_data_option
from microcosm_sagemaker.input_data import InputData


@command()
@input_data_option()
@input_artifact_option()
def main(input_data, input_artifact):
    graph = create_evaluate_app(
        loaders=[input_artifact.load_config],
    )

    run_evaluate(graph, input_data, input_artifact)


def run_evaluate(
    graph: ObjectGraph,
    input_data: InputData,
    input_artifact: InputArtifact,
) -> None:
    # Load the saved artifact
    graph.active_bundle.load(input_artifact)

    # Evaluate
    graph.active_evaluation(graph.active_bundle, input_data)
